/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_2.cpp

   Embedding samples
   Adding Default error handler.

   The default error handler will just print the string representation of
   intercepted errors on a provided Falcon Stream. If the user does not
   provide a stream, the error handler will crash; it is necessary to provide
   a minimal setup: Falcon provides a set of stdOutputStream(), stdErrorStream()
   and stdInputStream() object factories that will create a Falcon stream
   pointing to the named system resource.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_2.cpp -o falcon_embed_2

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}

// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   rt.loadFile( script_name );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   try {
      return myApp.embed( script_name );
   }
   catch( Falcon::Error* err )
   {
      // For now, let's use a simple STD:: provided stream.
      Falcon::AutoCString edesc( err->toString() );
      std::cerr << edesc.c_str() << std::endl;
      err->decref();
      return 1;
   }
}

/* end of falcon_embed_2.cpp */
