/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_5.cpp

   Embedding samples
   Adding userdata reflexivity for objects.

   This example shows how it is possible to provide C++ level data
   for scripts. It also shows reflexivity usage.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_5.cpp -o falcon_embed_5

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

//============================================================
// Begin of the userdata interface
//============================================================

// Our userdata will be used to store the count of times that
// refresh has been called and to reflect that info in the
// "times" property of the ProgramInfo Falcon class.

class ProgramInfo_obj: public Falcon::FalconObject
{
   int m_called;

public:

   // standard stuff
   ProgramInfo_obj( const Falcon::CoreClass* maker ):
      FalconObject( maker ),
      m_called( 0 )
   {}

   ProgramInfo_obj( const ProgramInfo_obj &other ):
      FalconObject( other ),
      m_called( other.m_called )
   {}

   virtual ~ProgramInfo_obj() {}

   // we reimplement here UserData interface
   virtual bool setProperty( const Falcon::String &prop, const Falcon::Item &value );
   virtual bool getProperty( const Falcon::String &key, Falcon::Item &ret ) const;
   virtual Falcon::CoreObject *clone() const;

   //===========================
   // Now a method to allow the code to change m_called.
   void incCalled() { m_called ++;}
};

// This is called each time the user wants to set a property
bool ProgramInfo_obj::getProperty( const Falcon::String &prop, Falcon::Item &value ) const
{
   if ( prop == "times" )
   {
      value = (Falcon::int64) m_called;
      return true;
   }
   return FalconObject::getProperty( prop, value );
}

bool ProgramInfo_obj::setProperty( const Falcon::String &prop, const Falcon::Item &value )
{
   if ( prop == "times" && value.isOrdinal() )
   {
      m_called = (int) value.forceInteger();
      return true;
   }
   return FalconObject::setProperty( prop, value );
}

// return a copy of ourselves, useful but not always necessary
Falcon::CoreObject *ProgramInfo_obj::clone() const
{
   return new ProgramInfo_obj( *this );
}

// The factory function is used by the falcon engine to create new
// instances as required by the script.
Falcon::CoreObject* ProgramInfoFactory( const Falcon::CoreClass* maker, void*, bool )
{
   return new ProgramInfo_obj( maker );
}

//============================================================
// Begin of the embedded module
//============================================================

// some forward decl
FALCON_FUNC say_hi( Falcon::VMachine *vm );

// methods of a class
FALCON_FUNC ProgramInfo_init( Falcon::VMachine *vm );
FALCON_FUNC ProgramInfo_refresh( Falcon::VMachine *vm );

Falcon::Module *createAppModule()
{
   // create a module and give it an unique name
   Falcon::Module *module = new Falcon::Module;
   module->name( "appModule" );

   // now add some constant...
   module->addConstant( "APP_REFRESH_FORWARD", (Falcon::int64) 0 );
   module->addConstant( "APP_REFRESH_BACKWARD", (Falcon::int64) 1 );

   // then add the functions
   module->addExtFunc( "say_hi", say_hi );

   // and add a class:
   // actually, adding the constructor (ProgramInfo_init) is optional
   Falcon::Symbol *cls_pinfo = module->addClass( "ProgramInfo", ProgramInfo_init );
   
   // Declare the object factory for this class.
   cls_pinfo->getClassDef()->factory( &ProgramInfoFactory );

   // add some property
   module->addClassProperty( cls_pinfo, "first" );
   module->addClassProperty( cls_pinfo, "second" );
   // add the reflected property
   module->addClassProperty( cls_pinfo, "times" );

   // and then add a method
   module->addClassMethod( cls_pinfo, "refresh", ProgramInfo_refresh );

   // the module is complete, we can return it
   return module;
}


FALCON_FUNC say_hi( Falcon::VMachine *vm )
{
   // how many parameters we have?
   Falcon::uint32 params = vm->paramCount();

   // raise an error if we have not enough params
   if ( params == 0 )
   {
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "Please, provide at least one parameter" ) ) );
      return;
   }

   // say hi to all parameters
   for ( Falcon::uint32 p = 0; p < params; p++ )
   {
      // Convert the item; instead of a simple conversion, involve the VM
      Falcon::String itmRep;
      vm->itemToString( itmRep, vm->param( p ) );

      vm->stdOut()->writeString( "Say hi to: " );
      vm->stdOut()->writeString( itmRep );
      vm->stdOut()->writeString( "\n" ); // \n is translated into the system line terminator
   }

   // return the number of items done
   vm->retval( (Falcon::int64) params );
}

// How to initialize a simple object
FALCON_FUNC ProgramInfo_init( Falcon::VMachine *vm )
{
   // let's get the parameters and store it in the properties
   Falcon::Item *p1 = vm->param( 0 ); // first param will go in "self.first"
   Falcon::Item *p2 = vm->param( 1 ); // second param will go in "self.second"

   // the self object is readied for us in the VM
   ProgramInfo_obj *self = Falcon::dyncast<ProgramInfo_obj *>( vm->self().asObject() );

   // has the first parameter been provided?
   if ( p1 != 0 )
      self->setProperty( "first", *p1 ); // will be a shallow copy
   else
      self->setProperty( "first", new Falcon::CoreString("Uninitialized") );

   if ( p2 != 0 )
      self->setProperty( "second", *p2 ); // will be a shallow copy
   else
      self->setProperty( "second", new Falcon::CoreString("Uninitialized") );

   // this is a constructor; there is no need to return anything.
}

// a simple method incrementing or decrementing numeric properties.
FALCON_FUNC ProgramInfo_refresh( Falcon::VMachine *vm )
{
   // this is a method. Get the self item.
   ProgramInfo_obj *self = Falcon::dyncast<ProgramInfo_obj *>( vm->self().asObject() );

   // we want one of our constants as a parameter
   Falcon::Item *refresh_mode = vm->param( 0 ); // first param will go in "self.first"

   if ( refresh_mode == 0 || ! refresh_mode->isOrdinal() )
   {
      throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "N" ) );  // standard falcon code for parameter declaration
   }

   int mode = (int) refresh_mode->forceInteger();
   if( mode != 0 && mode != 1 )
   {
      // standard falcon code for parameter declaration
      throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "must be an APP_REFRESH_* constant" ) );
      return;
   }

   Falcon::int64 returned = 0;

   Falcon::Item first;
   self->getProperty( "first", first );
   if ( first.isOrdinal() )
   {
      first.setInteger( first.forceInteger() + ( mode == 0 ? 1 : -1 ) );
      self->setProperty( "first", first );
      returned += first.asInteger();
   }

   Falcon::Item second;
   self->getProperty( "second", second );
   if ( second.isOrdinal() )
   {
      second.setInteger( second.forceInteger() + ( mode == 0 ? 1 : -1 ) );
      self->setProperty( "second", second );
      returned += second.asInteger();
   }

   // increment our data (as you see, no direct mangling of the "times" property here.
   self->incCalled();

   vm->retval( returned );
}

//============================================================
// End of the embedded module
//============================================================

//============================================================
// End of the embedded module
//============================================================
class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}

// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   rt.loadFile( script_name );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module

   // Finally add our application module
   Falcon::Module *appMod = createAppModule();
   vm->link( appMod );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm->findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm->findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm->findGlobalItem( "args" );

   // get the topmost (and so, the main) module, just to set the correct name.
   const Falcon::Module *mainMod = vm->mainModule()->module();

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::CoreString( mainMod->name() );
   *scriptPath = new Falcon::CoreString( script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray;
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::CoreString( argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. 
   Falcon::String str_regA;
   vm->regA().toString( str_regA );

   // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
   Falcon::AutoCString c_regA( str_regA );
   std::cout << "VM Output: " << c_regA.c_str() << std::endl;

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   // minimal housekeeping
   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
      return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream, 
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_5.cpp */
