/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_dynmod.cpp

   Embedding samples
   Dynamic modules.

   This example shows how to create modules that can create exported
   symbols on the fly instead providing them all at creation.

   The symbol "dynFunc" will not be added to the module at startup,
   but it will be made available when/if the script uses it.

   This code can be tested with dynImport.fal test.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_5.cpp -o falcon_embed_5

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

//============================================================
// Begin of the embedded module
//============================================================

// some forward decl
FALCON_FUNC say_hi( Falcon::VMachine *vm );
FALCON_FUNC dynFunc( Falcon::VMachine *vm );

// Our new dynamic module must provide a callback that will
// be called when the desired symbol is searched.

class MyDynModule: public Falcon::FlexyModule
{
public:
   virtual Falcon::Symbol *onSymbolRequest( const Falcon::String &name );
};

Falcon::Symbol* MyDynModule::onSymbolRequest( const Falcon::String &name )
{
   if( name == "dynFunc" )
   {
      return this->addExtFunc( "dynFunc", dynFunc );
   }

   // let the VM to solve its problem alone.
   return 0;
}

Falcon::Module *createAppModule()
{
   // create a module and give it an unique name
   Falcon::Module *module = new MyDynModule;
   module->name( "appDynModule" );

   // then add the functions
   module->addExtFunc( "say_hi", say_hi );

   // notice that we didn't add dynFunc; it will be added
   // only if the script requests it.

   // the module is complete, we can return it
   return module;
}


FALCON_FUNC say_hi( Falcon::VMachine *vm )
{
   // how many parameters we have?
   Falcon::uint32 params = vm->paramCount();

   // raise an error if we have not enough params
   if ( params == 0 )
   {
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "Please, provide at least one parameter" ) ) );
      return;
   }

   // say hi to all parameters
   for ( Falcon::uint32 p = 0; p < params; p++ )
   {
      // Convert the item; instead of a simple conversion, involve the VM
      Falcon::String itmRep;
      vm->itemToString( itmRep, vm->param( p ) );
      
      vm->stdOut()->writeString( "Say hi to: " );
      vm->stdOut()->writeString( itmRep );
      vm->stdOut()->writeString( "\n" ); // \n is translated into the system line terminator
   }

   // return the number of items done
   vm->retval( (Falcon::int64) params );
}

FALCON_FUNC dynFunc( Falcon::VMachine *vm )
{
   vm->stdOut()->writeString( "This function was dynamically loaded on explicit script request.\n" );
}

//============================================================
// End of the embedded module
//============================================================

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}


// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // Let's create a default error handler

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   rt.loadFile( script_name );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module

   // Finally add our application module
   Falcon::Module *appMod = createAppModule();
   vm->link( appMod );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm->findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm->findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm->findGlobalItem( "args" );

   // get the topmost (and so, the main) module, just to set the correct name.
   const Falcon::Module *mainMod = vm->mainModule()->module();

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::CoreString( mainMod->name() );
   *scriptPath = new Falcon::CoreString( script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray;
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::CoreString( argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. However, if the VM exited because
   // of runtime errors, the value in A should be ignored.
   Falcon::String str_regA;
   vm->regA().toString( str_regA );

   // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
   Falcon::AutoCString c_regA( str_regA );
   std::cout << "VM Output: " << c_regA.c_str() << std::endl;

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;
   
   // minimal housekeeping
   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream,
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
   
}

/* end of falcon_embed_5.cpp */
