/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_obj.cpp

   Embedding samples
   
   Overloading of core object class.

   In this example, we perform a complete overloading of the core object class,
   demonstrating the simplest and most powerful low-level embedding scheme.

   The extension object, called MyObject, can both contain application specific
   data (here we'll use a structure), or point to an application specific entity.

   The first case is called wrapper mode, as the object wraps the application data.
   
   In this second case, we talk of "carrier model", as the object acts as a simple
   carrier for application data.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) falcon_embed_obj.cpp -o falcon_embed_obj

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>
#include <string>

//============================================================
// Embedded Module -- class data
//============================================================

// a simple example of class data.
class AppData
{
public:
   std::string m_string;
   int m_value;

   AppData( const std::string &a_string, int a_value ):
      m_string( a_string ),
      m_value( a_value )
   {}

   AppData( const AppData& other ):
      m_string( other.m_string ),
      m_value( other.m_value )
   {
   }
};

//============================================================
// Embedded Module -- wrapper model.
//============================================================

class MyWrapper: public Falcon::CoreObject
{
   AppData m_data;
   
   // we demonstrate also how to cache explicitly Falcon items,
   // and how to keep falcon data. The std::string in data will be
   // reflected into a core class, but it's useless to create it
   // each time, so...
   mutable Falcon::CoreString* m_stringCache;

public:
   MyWrapper( const Falcon::CoreClass* generator, const std::string &str, int val );
   MyWrapper( const MyWrapper& );
   ~MyWrapper();
   
   virtual CoreObject *clone() const;
   virtual bool setProperty( const Falcon::String &prop, const Falcon::Item &value );
   virtual bool getProperty( const Falcon::String &key, Falcon::Item &ret ) const;
   virtual void gcMarkData( Falcon::uint32 mark );

   // you may also wish to do some automatic conversion between MyWrapper and AppData.
   // just for demonstration:
   void set( const AppData &app ) { m_data = app; }
   const AppData &get() const { return m_data; }
   AppData &get() { return m_data; }
};

MyWrapper::MyWrapper( const Falcon::CoreClass* generator, const std::string &str, int val ):
   Falcon::CoreObject( generator ),
   m_data( str, val ),
   m_stringCache(0)
{}

MyWrapper::MyWrapper( const MyWrapper& other ):
   CoreObject( other ),
   m_data( other.m_data ),
   m_stringCache(0)
{}

MyWrapper::~MyWrapper()
{
   // nothing to do; the string cache will be reclaimed by the GC.
}

Falcon::CoreObject *MyWrapper::clone() const
{
   // you may return 0 if the object cannot be cloned.
   return new MyWrapper( *this );
}

void MyWrapper::gcMarkData( Falcon::uint32 mark )
{
   // we must tell the string cache is safe with us
   if ( m_stringCache != 0 )
   {
      m_stringCache->mark( mark );
      // if we cache whole items instead, the correct procedure is to call
      // memPool->markItem( m_cached_item );
   }
}

bool MyWrapper::setProperty( const Falcon::String &prop, const Falcon::Item &value )
{
   if ( prop == "string" )
   {
      if ( ! value.isString() )
         return false; // or we may setup a more appropriate error message, see below.
         
      Falcon::AutoCString cstr( *value.asString() );
      m_data.m_string = cstr.c_str();
      return true;
   }
   else if ( prop == "number" )
   {
      if( ! value.isOrdinal() )
         throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
            .extra( "number is a numeric property" ) );
      
      m_data.m_value = (int) value.forceInteger();
      return true;
   }

   return false;  // prop not found.
}


bool MyWrapper::getProperty( const Falcon::String &prop, Falcon::Item &value ) const
{
   if ( prop == "string" )
   {
      // we must set value to a core string;
      // but creating a new string each time is boring... and inefficient, so...
      if ( m_stringCache == 0 )
      {
         m_stringCache = new Falcon::CoreString();
      }

      m_stringCache->bufferize( m_data.m_string.c_str() );
      value = m_stringCache;
      return true;
   }
   else if ( prop == "number" )
   {
      value = (Falcon::int64) m_data.m_value;
      return true;
   }
   else  {
      // Produce a mehtod given the symbol table we created in the module.
      return getMethodDefault(prop, value );
   }

   return false;
}

// Now a traditional method for us.
FALCON_FUNC MyWrapper_displayMe( Falcon::VMachine *vm )
{
   // dyncast is a dynamic_cast in debug, but is turned to a C cast in release.
   // vm->self() is a core object; in our case, a MyWrapper.
   MyWrapper* self = Falcon::dyncast<MyWrapper*>( vm->self().asObject() );
   std::cout << "Using a C++ stream - Content: " <<
      self->get().m_string << ",   value: " << self->get().m_value
      << std::endl;
}

// and if you plan to let the user create this objects in the script, you must
// provide a factory function that the engine is able to callback on object
// creation.
Falcon::CoreObject* MyWrapperFactory( const Falcon::CoreClass *cls, void *user_data, bool bDeserial )
{
   return new MyWrapper( cls, "Wrapped", 1 );
}

//============================================================
// Embedded Module -- carrier model.
//============================================================

class MyCarrier: public Falcon::CoreObject
{
   AppData* m_data;
   // skipping the string caching code.
   
public:
   MyCarrier( const Falcon::CoreClass* generator, AppData* data );
   MyCarrier( const MyCarrier& );
   ~MyCarrier();

   virtual CoreObject *clone() const;
   virtual bool setProperty( const Falcon::String &prop, const Falcon::Item &value );
   virtual bool getProperty( const Falcon::String &key, Falcon::Item &ret ) const;
   virtual void gcMarkData( Falcon::uint32 mark ) {}

   // don't check previous carried data for simpilcity here
   void set( AppData* ptr ) { m_data = ptr; }
   
   AppData* get() const { return m_data; }
};

MyCarrier::MyCarrier( const Falcon::CoreClass* generator, AppData* data ):
   Falcon::CoreObject( generator ),
   m_data( data )
{}

MyCarrier::MyCarrier( const MyCarrier& other ):
   CoreObject( other )
{
   m_data = new AppData( *other.get() );
   // you may wish to incremente a reference count instead.
   // In that case, be sure you have proper MT count protection if you may use
   // this item from more than one thread.
}

MyCarrier::~MyCarrier()
{
   delete m_data;
   // or you may wish to decrement a reference count.
}

Falcon::CoreObject *MyCarrier::clone() const
{
   // you may return 0 if the object cannot be cloned.
   return new MyCarrier( *this );
}


bool MyCarrier::setProperty( const Falcon::String &prop, const Falcon::Item &value )
{
   if ( prop == "string" )
   {
      if ( ! value.isString() )
         return false; // or we may setup a more appropriate error message, see below.

      Falcon::AutoCString cstr( *value.asString() );
      m_data->m_string = cstr.c_str();
      return true;
   }
   else if ( prop == "number" )
   {
      if( ! value.isOrdinal() )
         throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
            .extra( "number is a numeric property" ) );

      m_data->m_value = (int) value.forceInteger();
      return true;
   }

   return false;  // prop not found.
}


bool MyCarrier::getProperty( const Falcon::String &prop, Falcon::Item &value ) const
{
   if ( prop == "string" )
   {
      Falcon::CoreString* ret = new Falcon::CoreString( m_data->m_string.c_str() );
      ret->bufferize(); // be sure the data stays available.
      value = ret;
      return true;
   }
   else if ( prop == "number" )
   {
      value = (Falcon::int64) m_data->m_value;
      return true;
   }
   else  {
      // a method -- here we see the procedure in greater detail
      // We have stored methods in the property table.
      // We can use that map to retreive the available methods.
      const Falcon::Item* mth = generator()->properties().getValue( prop );
      if ( mth != 0 )
      {
         // yes, a valid method
         value = *mth;
         fassert( value.isFunction() ); // we know we have only methods.
         
         value.methodize( this );
         return true;
      }

      // for lightning fast perfomrance, you may overload the CoreClass
      // and record methods there, instead of having to search the property table.
   }

   return false;
}

// If your carrier class can be created by the script, you must provide an INIT:
FALCON_FUNC MyCarrier_init( Falcon::VMachine *vm )
{
   Falcon::Item* i_string = vm->param(0);
   Falcon::Item* i_value = vm->param(1);

   if( i_string == 0 || ! i_string->isString()
       || i_value == 0 || ! i_value->isOrdinal() )
   {
      throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
         .extra( "S,N" ) );
   }
   
   MyCarrier* self = Falcon::dyncast<MyCarrier*>( vm->self().asObject() );
   // the data is still not carried by the class, as the engine has created
   // an empty carrier with data 0.
   Falcon::AutoCString cstr( *i_string->asString() );
   self->set( new AppData( cstr.c_str(), i_value->forceInteger() ) );
}

// Now a traditional method for us.
FALCON_FUNC MyCarrier_displayMe( Falcon::VMachine *vm )
{
   // dyncast is a dynamic_cast in debug, but is turned to a C cast in release.
   // vm->self() is a core object; in our case, a MyWrapper.
   MyCarrier* self = Falcon::dyncast<MyCarrier*>( vm->self().asObject() );
   std::cout << "Using a C++ stream - Content: " <<
      self->get()->m_string << ",   value: " <<
      self->get()->m_value << std::endl;
}


Falcon::CoreObject* MyCarrierFactory( const Falcon::CoreClass *cls, void *user_data, bool bDeserial )
{
   // user_data may be zero; this happens when the factory is called by the engine
   // when the user creates an instance.
   return new MyCarrier( cls, static_cast<AppData *>(user_data) );
}

//============================================================
// Application Module
//============================================================

Falcon::Module* createAppModule()
{
   Falcon::Module* module = new Falcon::Module();

   // Create the wrapper class...
   Falcon::Symbol *cls_w = module->addClass( "MyWrapper" );

   // and assign the factory function
   cls_w->getClassDef()->factory( &MyWrapperFactory );

   // add the class propery table
   module->addClassProperty( cls_w, "string" );
   module->addClassProperty( cls_w, "number" );
   module->addClassMethod( cls_w, "displayMe", MyWrapper_displayMe );

   // similarly, create a Symbol for the carrier.
   Falcon::Symbol *cls_c = module->addClass( "MyCarrier", MyCarrier_init );

   // and assign the factory function
   cls_c->getClassDef()->factory( &MyCarrierFactory );

   // add the class propery table
   module->addClassProperty( cls_c, "string" );
   module->addClassProperty( cls_c, "number" );
   module->addClassMethod( cls_c, "displayMe", MyCarrier_displayMe );

   return module;
}

//============================================================
// End of the embedded module
//============================================================

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}


// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   rt.addModule( mainMod );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module
   // add also our module:
   Falcon::Module *appModule = createAppModule();
   vm->link( appModule );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // This is a shortcut setting the arguments and the standard script variables
   // in the main module; to be called when you're ready to launch
   vm->setupScript( argc, argv );

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We are done -- free the resources
   mainMod->decref();
   appModule->decref();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream,
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_obj.cpp */
