/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_sd.cpp

   Embedding samples
   Script data in overloaded (extended) VM

   At times, it is useful to bind some application data with
   a specific script. The VM has a hook called "userData" which
   can be used to store application or script specific data.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_evm->cpp -o falcon_embed_evm

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>
#include <list>
#include <string>

//==================================================
// Script specific data
//==================================================

class ScriptData
{
public:
   // this is just script-specific data that is used to show
   // how a local function can return some data which is related
   // to the running script
   int m_timesCalled;

   ScriptData():
      m_timesCalled( 0 )
   {}
};


//==================================================
// Application module module
//==================================================

// some forward decl
// This function will increment the call count in the script data
FALCON_FUNC appFunc1( Falcon::VMachine *vm );
FALCON_FUNC appFunc2( Falcon::VMachine *vm );
FALCON_FUNC appFunc3( Falcon::VMachine *vm );

// This function return the current call count, eventually resetting it
FALCON_FUNC callCount( Falcon::VMachine *vm );

Falcon::Module *createAppModule()
{
   // create a module and give it an unique name
   Falcon::Module *module = new Falcon::Module;
   module->name( "appModule" );

   // then add the functions
   module->addExtFunc( "appFunc1", appFunc1 );
   module->addExtFunc( "appFunc2", appFunc2 );
   module->addExtFunc( "appFunc3", appFunc3 );

   module->addExtFunc( "callCount", callCount );

   // the module is complete, we can return it
   return module;
}


FALCON_FUNC appFunc1( Falcon::VMachine *vm )
{
   //This function incremets the count time by one
   ScriptData *sd = reinterpret_cast<ScriptData *>(vm->userData());

   sd->m_timesCalled++;
}

FALCON_FUNC appFunc2( Falcon::VMachine *vm )
{
   //This function incremets the count time by two
   ScriptData *sd = reinterpret_cast<ScriptData *>(vm->userData());

   sd->m_timesCalled += 2;
}

FALCON_FUNC appFunc3( Falcon::VMachine *vm )
{
   //This function incremets the count time by three
   ScriptData *sd = reinterpret_cast<ScriptData *>(vm->userData());

   sd->m_timesCalled += 3;
}

FALCON_FUNC callCount( Falcon::VMachine *vm )
{
   ScriptData *sd = reinterpret_cast<ScriptData *>(vm->userData());

   // read the current script data value
   Falcon::int64 oldCount = sd->m_timesCalled;

   // If a parameter is given, we'll reset the counter to that value
   Falcon::Item *iNewCount = vm->param(0);

   if( iNewCount != 0 )
   {
      // also, if it's given, it should be a number...
      if ( ! iNewCount->isOrdinal() )
      {
         // standard falcon code for parameter declaration: N - Numeric
         vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
               extra( "N" ) ) );
         return;
      }
      
      // Ok, get the integer value of that item.
      sd->m_timesCalled = (int) iNewCount->forceInteger();
   }

   vm->retval( oldCount );
}


//============================================================
// End of the embedded module
//============================================================

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}


// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   rt.addModule( mainMod );

   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module
   // add also our module:
   Falcon::Module *appModule = createAppModule();
   vm->link( appModule );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // let's give our VM a valid Script Data.
   // Notice that the VM will just carry that data around;
   // it never disposes of it.
   ScriptData theScriptData;
   vm->userData( &theScriptData );

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm->findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm->findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm->findGlobalItem( "args" );

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::CoreString( mainMod->name() );
   *scriptPath = new Falcon::CoreString( script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray;
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::CoreString( argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();
   // great; let's see what's the final count of the script data
   std::cout << "At VM exit, Script data value: " << 
        theScriptData.m_timesCalled << std::endl;

   // We are done -- free the resources
   mainMod->decref();
   appModule->decref();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream, 
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_sd.cpp */
