/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_stream.cpp

   Embedding samples
   Overloading falcon streams.

   The best way to receive and manage script I/O is to overload falcon
   streams and provide the VM with them.

   We see also how to provide an UTF8 encoded stream as input (compiler)
   stream, in case your system default encoding is not UTF8.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_stream.cpp -o falcon_embed_stream

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-12 11:25:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
// We'll use system support to access an environment string
// System support is one of the thing you don't usually need in embedding,
// so it's not in falcon/engine.h
#include <falcon/sys.h>

#include <iostream>

//===============================================
// Reimplementation of standard Falcon streams
// We'll reimplement only output streams
//
// We'll implement a stream where every line
// is prefixed with "prefix:"
// So we need to buffer each line...
//===============================================

class MyStream: public Falcon::Stream
{
   const char *m_prefix;
   std::ostream *m_out;

   // This is where the buffer will be stored.
   Falcon::String m_buffer;

   // we're using this as a proxy for text writes.
   void internalWrite();

protected:
   virtual Falcon::int64 seek( Falcon::int64 pos, e_whence w );

public:
   MyStream( const char *prefix, std::ostream *out ):
      Stream( t_stream ),
      m_prefix( prefix ),
      m_out( out )
   {}

   // We don't really need to implement all of those;
   // as we want to reimplement output streams, we'll just
   // set "unsupported" where we don't want to provide support.
   bool writeString( const Falcon::String &source, Falcon::uint32 begin=0, Falcon::uint32 end = Falcon::csh::npos );
   virtual bool close();
   virtual Falcon::int32 read( void *buffer, Falcon::int32 size );
   virtual Falcon::int32 write( const void *buffer, Falcon::int32 size );
   virtual Falcon::int64 tell();
   virtual bool truncate( Falcon::int64 pos=-1 );
   virtual Falcon::int32 readAvailable( Falcon::int32 msecs_timeout );
   virtual Falcon::int32 writeAvailable( Falcon::int32 msecs_timeout );
   virtual Falcon::int64 lastError() const;
   virtual bool get( Falcon::uint32 &chr );
   virtual bool put( Falcon::uint32 chr );
};

// Byte oriented write.
// Scripts don't normally use standards VM streams to perform
// binary write, but there's no reason not to support it.
// We don't prefix nor translate, as this shall be some binary data
// as i.e. images.
Falcon::int32 MyStream::write( const void *buffer, Falcon::int32 size )
{
   const char *buf = (const char *)buffer;
   m_out->write( buf, size );
   return size;
}

// Text oriented write
bool MyStream::writeString( const Falcon::String &source, Falcon::uint32 begin, Falcon::uint32 end )
{
   if ( begin != 0 || end <= source.length() )
   {
      Falcon::String sub( source, begin, end );
      m_buffer += sub;
   }
   else
      m_buffer += source;

   internalWrite();

   return true;
}

// char oriented write (used mainly by transcoders)
bool MyStream::put( Falcon::uint32 chr )
{
   m_buffer += chr;
   internalWrite();

   return true;
}

Falcon::int64 MyStream::seek( Falcon::int64 pos, e_whence w )
{
   status( t_unsupported );
   return -1;
}


bool MyStream::close()
{
   return true;
}

Falcon::int32 MyStream::read( void *buffer, Falcon::int32 size )
{
   status( t_unsupported );
   return -1;
}

Falcon::int64 MyStream::tell()
{
   return 0;
}

bool MyStream::truncate( Falcon::int64 pos )
{
   status( t_unsupported );
   return false;
}

Falcon::int32 MyStream::readAvailable( Falcon::int32 msecs_timeout )
{
   status( t_unsupported );
   return -1;
}

Falcon::int32 MyStream::writeAvailable( Falcon::int32 msecs_timeout )
{
   return 1;
}

Falcon::int64 MyStream::lastError() const
{
   return 0;
}

bool MyStream::get( Falcon::uint32 &chr )
{
   status( t_unsupported );
   return false;
}


// this performs the real output
void MyStream::internalWrite()
{
   // we'll scan for each line in the buffer...
   Falcon::uint32 pos1 = 0;
   Falcon::uint32 pos2 = m_buffer.find( "\n" );

   while( pos2 != Falcon::String::npos )
   {
      // for each line, write from the start to the end (excluded)
      // we are wilfully changing Falcon "\n" with std::endl;
      // in Falcon, "\n" is considered the standard line terminator.
      // to expand it is duty of the embedder.

      Falcon::AutoCString toWrite( m_buffer.subString( pos1, pos2 ) );
      *m_out << m_prefix << toWrite.c_str() << std::endl;
      pos1 = pos2 + 1;
      pos2 = m_buffer.find( "\n", pos1 );
   }

   // trim the buffer
   m_buffer = m_buffer.subString( pos1 );
}


//===============================================
// End of reimplementation
//===============================================

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}


// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // it may be nice on us to detect a nonstardard source encoding asked by the user:
   Falcon::String srcEncoding;
   if ( Falcon::Sys::_getEnv( "FALCON_SRC_ENCODING", srcEncoding ) )
   {
      theLoader.sourceEncoding( srcEncoding );
   }
   else {
      // else, set UTF-8 encoding
      theLoader.sourceEncoding( "utf-8" );
   }

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   rt.loadFile( script_name );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   // change the VM streams that will be used by the scritps
   vm->stdOut( new MyStream( "Output: ", &std::cout ) );
   vm->stdErr( new MyStream( "Script Error: ", &std::cerr ) );

   vm->link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm->findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm->findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm->findGlobalItem( "args" );

   // get the topmost (and so, the main) module, just to set the correct name.
   const Falcon::Module *mainMod = vm->mainModule()->module();

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::CoreString( mainMod->name() );
   *scriptPath = new Falcon::CoreString( script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray;
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::CoreString( argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it.
   Falcon::String str_regA;
   vm->regA().toString( str_regA );

   // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
   Falcon::AutoCString c_regA( str_regA );
   std::cout << "VM Output: " << c_regA.c_str() << std::endl;
   
   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream, 
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_stream.cpp */
