/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_cb1.cpp

   Embedding samples
   VM callback support - predefined callback points

   This script shows how to perform callbacks on the script
   calling a routine "onLoad", then the main, then "onUnload".

   As it's also useful to have the module ready to inspect it
   in search for callback hooks, we take the occasion to show how
   to load the main module directly using the module loader.

   There are two ways to retreive the callback items:
   1) have the script to export them with the "export" directive and then
      load them in the embedding application with VMachine::findGlobalItem()
   2) Search them in the loaded module and get the item that the VM has
      generated to represent that symbol.

   With the first method a minimal cooperation is required on the script
   side, but this allow the script to define the hooks somewere else,
   in example, in loaded sub-modules, and export them from there.

   We're exploiting the first method here; scripts using callbacks
   will have to esport explicitly onLoad and onUnload functions.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_cb1.cpp -o falcon_embed_cb1

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}

// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   rt.addModule( mainMod );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm->findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm->findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm->findGlobalItem( "args" );

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::CoreString( mainMod->name() );
   *scriptPath = new Falcon::CoreString( script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray;
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::CoreString( argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   //======================================
   // find the callbacks in the main module
   //

   // our callbacks
   Falcon::GarbageLock *cbOnLoad = 0;
   Falcon::GarbageLock *cbOnUnload = 0;

   Falcon::Item *iOnLoad = vm->findGlobalItem( "onLoad" );
   // has the script defined an exported callable "onLoad" item?
   if ( iOnLoad != 0 && iOnLoad->isCallable() )
   {
      // Tell the GC that we're holding one of its internal items,
      // so that the GC won't free it if the script drops them.
      cbOnLoad = vm->lock( *iOnLoad );
   }

   // Do the same for onUnload
   Falcon::Item *iOnUnload = vm->findGlobalItem( "onUnload" );
   if ( iOnUnload != 0 && iOnUnload->isCallable() )
   {
      cbOnUnload = vm->lock( *iOnUnload );
   }

   //===================================
   // Launch the on load function
   //
   if ( cbOnLoad != 0 )
   {
      // we may use iOnLoad as well, but we show here the usage of
      // cbOnLoad as THAT is the object that should be stored safe
      // somewhere in a real program.

      // call without parameters
      vm->callItem( cbOnLoad->item(), 0 );

      // reset the event, that should be a normal "exit on return frame".
      vm->resetEvent();

      // we're not calling cbOnLoad anymore, so this is a good point to
      // allow the memory pool to dispose it at will
      vm->unlock( cbOnLoad );
   }

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. 
   Falcon::String str_regA;
   vm->regA().toString( str_regA );

   // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
   Falcon::AutoCString c_regA( str_regA );
   std::cout << "VM Output: " << c_regA.c_str() << std::endl;

   // calling onUnload with the returned item as a parameter.
   if( cbOnUnload != 0 )
   {
      // clear from the previous VM exit.
      vm->resetEvent();

      // push one parameter and call the function.
      vm->pushParameter( vm->regA() );
      vm->callItem( cbOnUnload->item(), 1 ); // one parameter.

      // we're not really interested to the output status of the VM

      // we're not calling cbOnUnload anymore, so this is a good point to
      // allow the VM to dispose it at will
      vm->unlock( cbOnUnload );
   }

   // We are done -- free the resources
   mainMod->decref();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream, 
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_cb1.cpp */
