/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_obj.cpp

   Embedding samples
   
   Exploiting of reflection models.

   Here we show how to use class reflection. Mainly, class (or property level) 
   reflection is performed through two devices: 
      - via reflective functions
      - via type/offset pairs
   
   In the first case, a reflection function is called back when the property 
   is read, and eventually set. The function can then set a return item to 
   the reflected value, or use the given falcon item to store the value in 
   the reflected item. 
   
   In the second case, the user_data stored inside the CoreObject (via the 
   setUserData() method) receives the C type specified in the property
   table, directly copied into it at an offset that was set when declaring 
   the class.
   
   This job is done by the ReflectObject and CRObject get/setProperty() 
   methods. The difference between the two classes is that the CRObject
   class also creates an item cache vector parallel to the property table. 
   That cache can be used either to store cached values (avoiding to create
   them in case this is heavy, or having them persistent if they reflect
   other deep objects in turn), and allowing the user to use non reflective
   properties as normal properties. The ReflectObject doesn't allocate a
   cache, and this means that it is not possible for the script users
   to alter properties unless explicitly reflected.
   
   You can try this binding with reflext_x.fal test.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) falcon_embed_obj.cpp -o falcon_embed_obj

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>
#include <string>

//============================================================
// Embedded Module -- application data
//============================================================

// a simple example of class data.
class AppData
{
public:
   std::string m_string;
   int m_value;

   AppData( const std::string &a_string, int a_value ):
      m_string( a_string ),
      m_value( a_value )
   {}

   AppData( const AppData& other ):
      m_string( other.m_string ),
      m_value( other.m_value )
   {
   }
};

// also, let's use a common base for the tedious work
class MyWrapperBase
{
protected:
   AppData m_data;
   
   MyWrapperBase( const std::string &str, int val );
   MyWrapperBase( const MyWrapperBase &other );
   ~MyWrapperBase();

public:
   // you may also wish to do some automatic conversion between MyWrapper and AppData.
   // just for demonstration:
   void set( const AppData &app ) { m_data = app; }
   const AppData &get() const { return m_data; }
   AppData &get() { return m_data; }

   static void reflectString_to(Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry& );
};

// And a simple constructor set.
MyWrapperBase::MyWrapperBase( const std::string &str, int val ):
   m_data( str, val )
{}

MyWrapperBase::MyWrapperBase( const MyWrapperBase &other ):
   m_data( other.m_data )
{}

MyWrapperBase::~MyWrapperBase()
{}

void MyWrapperBase::reflectString_to(Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry& )
{
   // you may decide to raise an error if this is not a string.
   if( ! property.isString() )
      throw new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ )
         .extra( "S" ) ); // we expected a string.
   
   Falcon::AutoCString newValue( *property.asString() );
   AppData *ad = reinterpret_cast<AppData*>( user_data );
   ad->m_string = newValue.c_str();
}

//============================================================
// Embedded Module -- Cached/Reflective wrapper.
//============================================================


class MyWrapperCached: public MyWrapperBase, public Falcon::CRObject
{
public:
   MyWrapperCached( const Falcon::CoreClass* generator, const std::string &str, int val );
   MyWrapperCached( const MyWrapperCached& );
   virtual ~MyWrapperCached();
   
   virtual CoreObject *clone() const;
   virtual void gcMarkData( Falcon::uint32 mark );
   
   static void reflectString_from( Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry& );
};

MyWrapperCached::MyWrapperCached( const Falcon::CoreClass* generator, const std::string &str, int val ):
   MyWrapperBase( str, val ),
   Falcon::CRObject( generator )
{
   // set the user data, so reflection knows where to take its data.
   setUserData( static_cast<void*>( &m_data ) );
}

MyWrapperCached::MyWrapperCached( const MyWrapperCached& other ):
   MyWrapperBase( other ),
   Falcon::CRObject( other )
{
   // set the user data, so reflection knows where to take its data.
   setUserData( static_cast<void*>( &m_data ) );
}

MyWrapperCached::~MyWrapperCached()
{
}

Falcon::CoreObject *MyWrapperCached::clone() const
{
   // you may return 0 if the object cannot be cloned.
   return new MyWrapperCached( *this );
}

void MyWrapperCached::gcMarkData( Falcon::uint32 mark )
{
   // nothing to do.
}

void MyWrapperCached::reflectString_from( Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry&pt )
{
   static Falcon::Item *cprop = 0;
   
   // we have a cache; exploit it
   Falcon::CacheObject* cco = Falcon::dyncast<Falcon::CacheObject *>( co );
   if( cprop == 0 )
   {
      cprop = cco->cachedProperty( *pt.m_name );
      // the other method may have initialized the data
      if ( cprop->isNil() )
         *cprop = new Falcon::CoreString();
   }
   
   AppData *ad = reinterpret_cast<AppData*>( user_data );
   // A finer algorithm may avoid re-buffering when m_string has not been changed.
   cprop->asString()->bufferize( ad->m_string.c_str() );
   
   property = *cprop; // will just perform a flat copy
}


// And the factory function.
Falcon::CoreObject* MyWrapperCachedFactory( const Falcon::CoreClass *cls, void *user_data, bool bDeserial )
{
   return new MyWrapperCached( cls, "Wrapped", 1 );
}

//============================================================
// Embedded Module -- Purely Reflective wrapper.
//============================================================

class MyWrapperUncached: public MyWrapperBase, public Falcon::ReflectObject
{
public:
   MyWrapperUncached( const Falcon::CoreClass* generator, const std::string &str, int val );
   MyWrapperUncached( const MyWrapperUncached& );
   virtual ~MyWrapperUncached();
   
   virtual CoreObject *clone() const;
   virtual void gcMarkData( Falcon::uint32 mark );

   static void reflectString_from( Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry& );
};

MyWrapperUncached::MyWrapperUncached( const Falcon::CoreClass* generator, const std::string &str, int val ):
   MyWrapperBase( str, val ),
   Falcon::ReflectObject( generator, static_cast<void*>( &m_data ) )
{}

MyWrapperUncached::MyWrapperUncached( const MyWrapperUncached& other ):
   MyWrapperBase( other ),
   Falcon::ReflectObject( other )
{}

MyWrapperUncached::~MyWrapperUncached()
{
}

Falcon::CoreObject *MyWrapperUncached::clone() const
{
   // you may return 0 if the object cannot be cloned.
   return new MyWrapperUncached( *this );
}

void MyWrapperUncached::gcMarkData( Falcon::uint32 mark )
{
   // nothing to do.
}


void MyWrapperUncached::reflectString_from( Falcon::CoreObject *co, void *user_data, Falcon::Item &property, const Falcon::PropEntry&pt )
{
   AppData *ad = reinterpret_cast<AppData*>( user_data );
   Falcon::CoreString *str = new Falcon::CoreString;
   str->bufferize( ad->m_string.c_str() );
   property = str;
}


// and if you plan to let the user create this objects in the script, you must
// provide a factory function that the engine is able to callback on object
// creation.
Falcon::CoreObject* MyWrapperUncachedFactory( const Falcon::CoreClass *cls, void *user_data, bool bDeserial )
{
   return new MyWrapperUncached( cls, "Wrapped", 1 );
}

//============================================================
// A common method -- using traditional access to self()
//============================================================

FALCON_FUNC MyWrapper_displayMe( Falcon::VMachine *vm )
{
   // We set our data as if it was carried, so it doesn't matter
   // which subclass are we in.
   AppData* self_data = reinterpret_cast<AppData*>( vm->self().asObject()->getUserData() );
   std::cout << "Using a C++ stream - Content: " <<
      self_data->m_string << ",   value: " << self_data->m_value
      << std::endl;
}

//============================================================
// Application Module
//============================================================

Falcon::Module* createAppModule()
{
   Falcon::Module* module = new Falcon::Module();

   // Create the wrapper class...
   Falcon::Symbol *cls_w = module->addClass( "MyWrapperCached" );

   // and assign the factory function
   cls_w->getClassDef()->factory( &MyWrapperCachedFactory );

   // add the class propery table -- set reflection via functions.
   module->addClassProperty( cls_w, "string" )
      .setReflectFunc( &MyWrapperCached::reflectString_from, &MyWrapperBase::reflectString_to );
   
   // automatic reflection via data type/offset displacement.
   AppData dummy_app_data( "dummy", 0 );
   module->addClassProperty( cls_w, "number" )
      .setReflective( Falcon::e_reflectInt, &dummy_app_data, &dummy_app_data.m_value );

   // finally, a standard method.
   module->addClassMethod( cls_w, "displayMe", MyWrapper_displayMe );

   //=================================================================
   // show that we can do that also without a cache.
   Falcon::Symbol *cls_wu = module->addClass( "MyWrapperUncached" );
   cls_wu->getClassDef()->factory( &MyWrapperUncachedFactory );

   module->addClassProperty( cls_wu, "string" )
      .setReflectFunc( &MyWrapperUncached::reflectString_from, &MyWrapperBase::reflectString_to );

   module->addClassProperty( cls_wu, "number" )
      .setReflective( Falcon::e_reflectInt, &dummy_app_data, &dummy_app_data.m_value );

   // The trick is to set the methods as read-only, 
   // so that they cannot be changed into something else, and then need no cache.
   module->addClassMethod( cls_wu, "displayMe", MyWrapper_displayMe )
      .setReadOnly( true );

   return module;
}

//============================================================
// End of the embedded module
//============================================================

class AppFalcon
{
public:
   AppFalcon();
   ~AppFalcon();
   int embed( const char *script_name, int argc, char **argv );
};

AppFalcon::AppFalcon()
{
   Falcon::Engine::Init();
}

AppFalcon::~AppFalcon()
{
   Falcon::Engine::Shutdown();
}


// This is the routine that embeds falcon
int AppFalcon::embed( const char *script_name, int argc, char **argv )
{
   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::ModuleLoader theLoader(".");

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   rt.addModule( mainMod );

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachineWrapper vm;

   vm->link( Falcon::core_module_init() );  // add the core module
   // add also our module:
   Falcon::Module *appModule = createAppModule();
   vm->link( appModule );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   vm->link( &rt );

   // This is a shortcut setting the arguments and the standard script variables
   // in the main module; to be called when you're ready to launch
   vm->setupScript( argc, argv );

   // we're ready to go. Still, we may fail if the script has not a main routine.
   vm->launch();

   // We are done -- free the resources
   mainMod->decref();
   appModule->decref();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   AppFalcon myApp;

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   try {
     return myApp.embed( script_name, argc - 2, argv + 2 );
   }
   catch( Falcon::Error* err )
   {
      // This time let's use a Falcon stream,
      // that knows how to handle Falcon strings.
      Falcon::Stream* stdErr = new Falcon::StdErrStream();
      stdErr->writeString( err->toString() );
      err->decref();
      delete stdErr;
   }
}

/* end of falcon_embed_obj.cpp */
